(* icosactlr.ml     Controller for IcosaBlue example
 *
 * Copyright (c) 2011 Psellos   http://psellos.com
 * Licensed under the MIT license:
 *     http://www.opensource.org/licenses/mit-license.php
 *)

let facect = 20    (* Number of faces of icosahedron *)
let sidect = 3     (* Number of sides of each face (triangle) *)

let txfullwd = 512 (* Full texture width (two triangle images) *)
let txwd = 256     (* Texture width (one triangle) *)
let txht = 256     (* Texture height *)

let ftxfullwd = float_of_int txfullwd
let ftxwd = float_of_int txwd
let ftxht = float_of_int txht

(* Geometry of texture image.
 *)
let s3o2 = sqrt 3.0 /. 2.0   (* Equilateral triangle constant *)
let imgside = 189.0          (* Triangle side *)
let imgheight =              (* Triangle height *)
    Cocoa.frnd (imgside *. s3o2)
let inmar = 20.0             (* Margin of inner triangle *)

(* Colors.
 *)
let dark_blue = (0.153, 0.263, 0.733, 1.0)
let bright_blue = (0.125, 0.196, 0.918, 1.0)
let beige = (0.882, 0.824, 0.796, 1.0)


let draw_triangle cgctx xorigin =
    (* Draw the triangular texture image for a side of the icosahedron.
     * There's a background color and an inner triangle with rounded
     * corners.
     *)
    let s = imgside  (* Convenient abbreviation of image side*)
    in let i = inmar (* Convenient abbreviation of indent *)
    in let t = 0.1   (* Trimmed distance, beginning of rounded corner *)
    in let c = 0.05  (* Control distance, position of Bezier control point *)
    in let segs = [|
        (* Segments making up the inner triangle.
         *)
        ((i *. s3o2, i /. 2.0), (s /. 2.0, s *. s3o2 -. i));
        ((s /. 2.0, s *. s3o2 -. i), (s -. i *. s3o2, i /. 2.0));
        ((s -. i *. s3o2, i /. 2.0), (i *. s3o2, i /. 2.0));
    |]
    in let otrans ((x1, y1), (x2, y2)) =
        ((xorigin +. x1, y1), (xorigin +. x2, y2))
    in let osegs = Array.map otrans segs
    in let trimseg f ((x1, y1), (x2, y2)) =
        ((x1 +. f *. (x2 -. x1),
          y1 +. f *. (y2 -. y1)),
         (x1 +. (1.0 -. f) *. (x2 -. x1),
         y1 +. (1.0 -. f) *. (y2 -. y1)))
    in let tsegs = Array.map (trimseg t) osegs
    in let csegs = Array.map (trimseg c) osegs
    in let add_side (_, p) (_, c1) (c2, _) (q, _) =
        begin
        CgContext.addLineToPoint cgctx p;
        CgContext.addCurveToPoint cgctx c1 c2 q;
        end
    in
        begin
        CgContext.setFillColor cgctx dark_blue;
        CgContext.fillRect cgctx (xorigin, 0.0, ftxwd, ftxht);
        CgContext.moveToPoint cgctx (fst tsegs.(0));
        for i = 0 to 2 do
            let i' = (i + 1) mod 3
            in
                add_side tsegs.(i) csegs.(i) csegs.(i') tsegs.(i')
        done;
        CgContext.setFillColor cgctx bright_blue;
        CgContext.fillPath cgctx;
        end


let text_width cgctx text =
    (* Calculate the width of the text using current settings of the
     * CgContext.  Caller must beware that the text drawing mode will be
     * changed by this call.
     *)
    let () =
        begin
        CgContext.setTextPosition cgctx (0.0, 0.0);
        CgContext.setTextDrawingMode cgctx CgContext.Invisible;
        CgContext.showText cgctx text (String.length text);
        end
    in
        fst (CgContext.getTextPosition cgctx)


let draw_psellos cgctx xorigin =
    (* Draw a Greek Psi and Psellos.com into the triangle for "branding"
     * purposes.  Forgive the self-promotion.
     *)
    let psimar = 17.0   (* Extra margin around the Psi *)
    in let txbase = 8.0 (* Baseline for text *)
    in let text = "Psellos.com"
    in let rect =
        (xorigin +. inmar +. psimar,
         inmar +. psimar,
         imgside -. 2.0 *. inmar -. 2.0 *. psimar,
         imgheight -. 3.0 *. inmar -. 2.0 *. psimar)
    in let sfig = Bzpdata.bzp_scale Bzpdata.figure1 rect
    in let drawit = function
        | Bzpdata.BZBegin -> CgContext.beginPath cgctx
        | Bzpdata.BZMove p -> CgContext.moveToPoint cgctx p
        | Bzpdata.BZLine p -> CgContext.addLineToPoint cgctx p
        | Bzpdata.BZCurve (c, d, p) -> CgContext.addCurveToPoint cgctx c d p
        | Bzpdata.BZClose -> CgContext.closePath cgctx
    in
        begin
        (* First the Greek Psi.
         *)
        Bzpdata.bzp_iter drawit sfig;
        CgContext.setFillColor cgctx beige;
        CgContext.fillPath cgctx;

        (* Now the text.
         *)
        CgContext.selectFont cgctx "Verdana" 16.0;
        let indent = floor ((imgside -. text_width cgctx text) /. 2.0)
        in let pos = (xorigin +. indent, inmar +. txbase)
        in
            CgContext.setTextPosition cgctx pos;
        CgContext.setTextDrawingMode cgctx CgContext.FillStroke;
        CgContext.setFillColor cgctx beige;
        CgContext.setStrokeColor cgctx beige;
        CgContext.showText cgctx text (String.length text);
        end


let select_texture faceix txix =
    (* Select the given texture for the given face.  Right now there are
     * only two textures, numbered 0 and 1.  The default is 0, which is
     * plain.  Texture 1 has the "branding".
     *)
    let xorigin = (float_of_int txix *. ftxwd) /. ftxfullwd
    in
        for i = 0 to 2 do
            let vix = Raw.get Icosamodel.faces ~pos: (3 * faceix + i)
            in let xold = Raw.get_float Icosamodel.texcos ~pos: (2 * vix)
            in let xoffset = mod_float xold (ftxwd /. ftxfullwd)
            in
                Raw.set_float Icosamodel.texcos ~pos: (2 * vix)
                    (xorigin +. xoffset)
        done


class t =
object (self)
    inherit Wrappee.t

    val interval = 1.0 /. 24.0  (* Interval between timer ticks *)

    val mutable active : bool = true
    val mutable ticks = 0       (* Number of ticks since inception *)

    (* View that delegates to us.
     *)
    val mutable theDelegator: UiView.t option = None

    (* OpenGL state.
     *)
    val mutable theGlctx: EAGLContext.t option = None
    val mutable view_renderbuffer = GlFramebuffer.renderbuffer0
    val mutable view_depthbuffer = GlFramebuffer.renderbuffer0
    val mutable view_framebuffer = GlFramebuffer.framebuffer0
    val mutable backing_width = 0
    val mutable backing_height = 0

    (* Textures.
     *)
    val mutable theCgctx : CgContext.t option = None
    val mutable thePixmap : ([`rgba], [`ubyte]) GlPix.t option = None

    (* Accessors.
     *)
    method delegator =
        match theDelegator with
        | None -> failwith "Initialization error: no delegator (IcosaBlue.nib)"
        | Some view -> view
    
    method setDelegator' view =
        theDelegator <- Some view

    method glctx =
        match theGlctx with
        | None -> failwith "Initialization error: no OpenGL context"
        | Some c -> c

    method set_glctx glctx =
        theGlctx <- Some glctx

    method cgctx =
        match theCgctx with
        | None -> failwith "Initialization error: no CgContext"
        | Some c -> c

    method set_cgctx c =
        theCgctx <- Some c

    method pixmap =
        match thePixmap with
        | None -> failwith "Initialization error: no GlPix"
        | Some p -> p

    method set_pixmap p =
        thePixmap <- Some p

    (* Application state change events.
     *)
    method applicationDidFinishLaunching' (app: UiApplication.t) =
        (* Startup code.
         *)
        begin
        (UiApplication.sharedApplication ())#setStatusBarStyle'
            UiApplication.StatusBarStyleBlackOpaque;
        self#delegator#layer#setOpaque' true;
        self#set_glctx
            ((EAGLContext.alloc ())#initWithAPI'
                    EAGLContext.RenderingAPIOpenGLES1);
        self#set_cgctx (CgContext.create txfullwd txht);
        ignore (
            NsTimer.scheduledTimer5 interval self "drawView'" Wrappee.nil true
        );

        (* Set coordinate systems of texture graphics context to
         * match the OpenGL coordinate system.
         *)
        CgContext.scaleCTM self#cgctx 1.0 (-1.0);
        CgContext.translateCTM self#cgctx 0.0 (-. (float_of_int txht));
        CgContext.setTextMatrix self#cgctx (1.0, 0.0, 0.0, -1.0, 0.0, 0.0);

        (* Draw textures and select textures for faces.
         *)
        draw_triangle self#cgctx 0.0;
        draw_triangle self#cgctx ftxwd;
        draw_psellos self#cgctx ftxwd;
        self#set_pixmap
            (GlPix.of_raw (CgContext.getData self#cgctx)
                ~format: `rgba ~width: txfullwd ~height: txht);

        select_texture 4 1;  (* Just one face has the "branding". *)
        end

    method applicationDidEnterBackground' (app: UiApplication.t) =
        active <- false

    method applicationWillEnterForeground' (app: UiApplication.t) =
        active <- true

    method applicationDidReceiveMemoryWarning' (app: UiApplication.t) =
        Gc.compact () (* Best effort to reclaim space *)

    (* OpenGL drawing.
     *)
            
    method viewLayoutSubviews' (view: UiView.t) =
        (* This method is called at the beginning, and after changes to
         * the geometry of the view.  (Re)allocate and (re)initialize
         * the framebuffer state.
         *)
        begin
        ignore self#make_current;
        self#destroyFramebuffer;
        ignore self#createFramebuffer;

        Gl.enable `depth_test;

        (* Texture state.
         *)
        List.iter (GlTex.parameter ~target: `texture_2d) [
            (`wrap_s `clamp_to_edge);
            (`wrap_t `clamp_to_edge);
            (`mag_filter `linear);
            (`min_filter `linear);
        ];
        Gl.enable `texture_2d;
        GlTex.env (`mode `modulate);
        GlTex.image2d ~internal: `rgba self#pixmap;

        (* Perspective projection.
         *)
        GlMat.mode `projection;
        let near, far, fieldofview = 0.01, 1000.0, atan 1.0
        in let size = near *. tan (fieldofview /. 2.0)
        in let rat = float_of_int backing_height /. float_of_int backing_width
        in
            GlMat.frustum
                ~x: (-. size, size)
                ~y: (-. size *. rat, size *. rat)
                ~z: (near, far);
        GlDraw.viewport ~x: 0 ~y: 0 ~w: backing_width ~h: backing_height;

        (* Lighting.
         *)
        Gl.enable `lighting;
        Gl.enable `light0;
        GlLight.light 0 (`ambient (0.3, 0.3, 0.3, 1.0));
        GlLight.light 0 (`diffuse (0.6, 0.6, 0.6, 1.0));
        GlLight.light 0 (`specular (0.5, 0.5, 0.5, 1.0));
        GlLight.light 0 (`position (10.0, 10.0, 10.0, 1.0));
        GlLight.light 0 (`spot_direction (-1.0, -1.0, -1.0));
        GlLight.light 0 (`spot_cutoff 20.0);

        GlLight.material ~face: `both
            (`ambient_and_diffuse (1.0, 1.0, 1.0, 1.0));
        GlLight.material ~face: `both
            (`specular (0.3, 0.3, 0.3, 1.0));
        GlLight.material ~face: `both
            (`shininess 30.0);

        (* Model/view matrix is continually updated to do animation.
         *)
        GlMat.mode `modelview;

        self#drawView' (Obj.magic Wrappee.nil);
        end

    method drawView' (timer: NsTimer.t) =
        (* Display one frame of the icosahedron.
         *)
        if active then
            begin
            ignore self#make_current;
            GlFramebuffer.bind_framebuffer `framebuffer view_framebuffer;

            (* Each frame rotates a little and displays the new view.
             * To keep things slightly more interesting, the axis of
             * rotation moves in a circle.
             *)
            GlMat.load_identity ();
            GlMat.translate ~z: (-3.0) ();
            let angle = -2.0 *. float_of_int ticks
            in let precession = float_of_int ticks *. 0.00314159265358979324
            in let inclination = 0.2
            in let x = inclination *. cos precession
            in let z = inclination *. sin precession
            in
                GlMat.rotate ~angle ~x ~y: 1.0 ~z ();
            GlMat.rotate ~angle: 58.282525588538995 ~x: 1.0 ();

            ticks <- ticks + 1;

            GlClear.color (0.0, 0.0, 0.0);
            GlClear.clear [ `color ; `depth ];
            GlArray.enable `vertex;
            GlArray.enable `normal;
            GlArray.enable `texture_coord;
            GlArray.vertex `three Icosamodel.vertices;
            GlArray.normal Icosamodel.normals;
            GlArray.tex_coord `two Icosamodel.texcos;
            GlArray.draw_elements
                `triangles ~count: (sidect * facect) Icosamodel.faces;
            GlArray.disable `vertex;
            GlArray.disable `normal;
            GlArray.disable `texture_coord;
            
            GlFramebuffer.bind_renderbuffer `renderbuffer view_renderbuffer;
            self#present_buffer;
            end

    method private make_current : bool =
        EAGLContext.setCurrentContext' self#glctx

    method private buffer_storage : bool =
        self#glctx#renderbufferStorage'fromDrawable'
            `renderbuffer self#delegator#layer

    method private present_buffer =
        ignore (self#glctx#presentRenderbuffer' `renderbuffer)
    
    method private createFramebuffer =
        let () =
            begin
            view_framebuffer <- GlFramebuffer.gen_framebuffer ();
            view_renderbuffer <- GlFramebuffer.gen_renderbuffer ();
            GlFramebuffer.bind_framebuffer `framebuffer view_framebuffer;
            GlFramebuffer.bind_renderbuffer `renderbuffer view_renderbuffer;
            ignore self#buffer_storage;
            GlFramebuffer.renderbuffer
                `framebuffer `color0 `renderbuffer view_renderbuffer;
            backing_width <-
                GlFramebuffer.get_renderbuffer_parameter `renderbuffer `width;
            backing_height <-
                GlFramebuffer.get_renderbuffer_parameter `renderbuffer `height;
            (* Depth buffer
             *)
            view_depthbuffer <- GlFramebuffer.gen_renderbuffer ();
            GlFramebuffer.bind_renderbuffer `renderbuffer view_depthbuffer;
            GlFramebuffer.renderbuffer_storage
                `renderbuffer `depth_component16 backing_width backing_height;
            GlFramebuffer.renderbuffer
                `framebuffer `depth `renderbuffer view_depthbuffer;
            end
        in
            GlFramebuffer.check_status `framebuffer = `complete
        
    method private destroyFramebuffer =
        begin
        GlFramebuffer.delete_framebuffer view_framebuffer;
        view_framebuffer <- GlFramebuffer.framebuffer0;
        GlFramebuffer.delete_renderbuffer view_renderbuffer;
        view_renderbuffer <- GlFramebuffer.renderbuffer0;
        GlFramebuffer.delete_renderbuffer view_depthbuffer;
        view_depthbuffer <- GlFramebuffer.renderbuffer0;
        end
end


let wrapped robjcv =
    let c = new t in c#setContainer robjcv; c
in
    Callback.register "Icosactlr.wrapped" wrapped
