/* $Id: ml_gl.c,v 1.51 2007/04/13 02:48:43 garrigue Exp $ */

#ifdef _WIN32
#include <wtypes.h>
#endif
#include <string.h>
#ifdef __APPLE__
#include <OpenGLES/ES1/gl.h>
#include <OpenGLES/ES1/glext.h>
#else
#include <GLES/gl.h>
#endif
#ifdef HAS_GLEXT_H
#include <GLES/glext.h>
#undef GL_VERSION_1_3
#endif
#include <caml/misc.h>
#include <caml/mlvalues.h>
#include <caml/callback.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/fail.h>
#include "ml_raw.h"
#include "gl_tags.h"
#include "ml_gl.h"


void ml_raise_gl(const char *errmsg)
{
  static value * gl_exn = NULL;
  if (gl_exn == NULL)
      gl_exn = caml_named_value("glerror");
  raise_with_string(*gl_exn, (char*)errmsg);
}

value copy_string_check (const unsigned char *str)
{
    if (!str) ml_raise_gl("Null string");
    return copy_string ((char*) str);
}

value copy_string_lenient(const unsigned char *str)
{
    return copy_string (str == NULL ? "" : (char *) str);
}

struct record {
    value key; 
    GLenum data;
};

static struct record input_table[] = {
#include "gl_tags.c"
};

static struct record *tag_table = NULL;

#define TABLE_SIZE (TAG_NUMBER*2+1)

CAMLprim value ml_gl_make_table (value unit)
{
    int i;
    unsigned int hash;

    tag_table = stat_alloc (TABLE_SIZE * sizeof(struct record));
    memset ((char *) tag_table, 0, TABLE_SIZE * sizeof(struct record));
    for (i = 0; i < TAG_NUMBER; i++) {
	hash = (unsigned long) input_table[i].key % TABLE_SIZE;
	while (tag_table[hash].key != 0) {
	    hash ++;
	    if (hash == TABLE_SIZE) hash = 0;
	}
	tag_table[hash].key = input_table[i].key;
	tag_table[hash].data = input_table[i].data;
    }
    return Val_unit;
}

GLenum GLenum_val(value tag)
{
    unsigned int hash = (unsigned long) tag % TABLE_SIZE;

    if (!tag_table) ml_gl_make_table (Val_unit);
    while (tag_table[hash].key != tag) {
	if (tag_table[hash].key == 0) ml_raise_gl ("Unknown tag");
	hash++;
	if (hash == TABLE_SIZE) hash = 0;
    }
    /*
    fprintf(stderr, "Converted %ld to %d", Int_val(tag), tag_table[hash].data);
    */
    return tag_table[hash].data;
}

/*
GLenum GLenum_val(value tag)
{
    switch(tag)
    {
#include "gl_tags.c"
    }
    ml_raise_gl("Unknown tag");
}
*/

ML_2 (glAlphaFunc, GLenum_val, Float_val)

ML_2 (glBlendFunc, GLenum_val, GLenum_val)

CAMLprim value ml_glClipPlanef(value plane, value equation)  /* ML */
{
    GLfloat eq[4];
    int i;

    for (i = 0; i < 4; i++)
	eq[i] = Float_val (Field(equation,i));
    glClipPlanef (GL_CLIP_PLANE0 + Int_val(plane), eq);
    return Val_unit;
}

CAMLprim value ml_glClear(value bit_list)  /* ML */
{
    GLbitfield accu = 0;

    while (bit_list != Val_int(0)) {
	switch (Field (bit_list, 0)) {
	case MLTAG_color:
	    accu |= GL_COLOR_BUFFER_BIT; break;
	case MLTAG_depth:
	    accu |= GL_DEPTH_BUFFER_BIT; break;
	case MLTAG_stencil:
	    accu |= GL_STENCIL_BUFFER_BIT; break;
	}
	bit_list = Field (bit_list, 1);
    }
    glClear (accu);
    return Val_unit;
}
ML_4 (glClearColor, Float_val, Float_val, Float_val, Float_val)
ML_1 (glClearDepthf, Float_val)
ML_1 (glClearStencil, Int_val)
ML_4 (glColor4f, Float_val, Float_val, Float_val, Float_val)
ML_4 (glColorMask, Int_val, Int_val, Int_val, Int_val)
ML_1 (glCullFace, GLenum_val)

ML_1 (glDisable, GLenum_val)
ML_1 (glDepthFunc, GLenum_val)
ML_1 (glDepthMask, Int_val)
ML_2 (glDepthRangef, Float_val, Float_val)

ML_1 (glEnable, GLenum_val)

CAMLprim value ml_glFog (value param) /* ML */
{
    GLfloat params[4];
    int i;

    switch (Field(param,0))
    {
    case MLTAG_mode:
	glFogf(GL_FOG_MODE, (GLfloat) GLenum_val(Field(param,1)));
	break;
    case MLTAG_density:
	glFogf(GL_FOG_DENSITY, Float_val(Field(param,1)));
	break;
    case MLTAG_start:
	glFogf(GL_FOG_START, Float_val(Field(param,1)));
	break;
    case MLTAG_End:
	glFogf(GL_FOG_END, Float_val(Field(param,1)));
	break;
    case MLTAG_color:
      for (i = 0; i < 4; i++) params[i] = Float_val(Field(Field(param,1),i));
	glFogfv(GL_FOG_COLOR, params);
	break;
    }
    return Val_unit;
}

ML_0 (glFlush)
ML_0 (glFinish)
ML_1 (glFrontFace, GLenum_val)
ML_3 (glFrustumf, Pair(arg1,Float_val,Float_val),
      Pair(arg2,Float_val,Float_val), Pair(arg3,Float_val,Float_val))

#ifdef CONFIG_IOSSIM
/* When using the iPhone Simulator from the command line, glGetString()
 * has been known to return NULL when things are otherwise working well.
 * So treat a NULL return leniently.
 */
ML_1_ (glGetString, GLenum_val, copy_string_lenient)
#else
ML_1_ (glGetString, GLenum_val, copy_string_check)
#endif

ML_2 (glGetFloatv, GLenum_val, Float_raw)

CAMLprim value ml_glGetError(value unit)
{
    switch (glGetError()) {
    case GL_NO_ERROR:       return MLTAG_no_error;
    case GL_INVALID_ENUM:   return MLTAG_invalid_enum;
    case GL_INVALID_VALUE:  return MLTAG_invalid_value;
    case GL_INVALID_OPERATION:  return MLTAG_invalid_operation;
    case GL_STACK_OVERFLOW: return MLTAG_stack_overflow;
    case GL_STACK_UNDERFLOW: return MLTAG_stack_underflow;
    case GL_OUT_OF_MEMORY:  return MLTAG_out_of_memory;
    default: ml_raise_gl("glGetError: unknown error");
    }
}
	
CAMLprim value ml_glHint (value target, value hint)
{
    GLenum targ = 0U;

    switch (target) {
    case MLTAG_fog:	targ = GL_FOG_HINT; break;
    case MLTAG_line_smooth:	targ = GL_LINE_SMOOTH_HINT; break;
    case MLTAG_perspective_correction:
	targ = GL_PERSPECTIVE_CORRECTION_HINT; break;
    case MLTAG_point_smooth:	targ = GL_POINT_SMOOTH_HINT; break;
    }
    glHint (targ, GLenum_val(hint));
    return Val_unit;
}

ML_1_ (glIsEnabled, GLenum_val, Val_int)

CAMLprim value ml_glLight (value n, value param)  /* ML */
{
    float params[4];
    int i;

    if (Int_val(n) >= GL_MAX_LIGHTS) invalid_argument ("Gl.light");
    switch (Field(param,0))
    {
    case MLTAG_ambient:
    case MLTAG_diffuse:
    case MLTAG_specular:
    case MLTAG_position:
	for (i = 0; i < 4; i++)
	    params[i] = Float_val (Field(Field(param, 1), i));
	break;
    case MLTAG_spot_direction:
	for (i = 0; i < 3; i++)
	    params[i] = Float_val (Field(Field(param, 1), i));
	break;
    default:
	params[0] = Float_val (Field(param, 1));
    }
    glLightfv (GL_LIGHT0 + Int_val(n), GLenum_val(Field(param,0)), params);
    return Val_unit;
}

CAMLprim value ml_glLightModel (value param)  /* ML */
{
    GLfloat params[4];
    int i;

    switch (Field(param,0))
    {
    case MLTAG_ambient:
	for (i = 0; i < 4; i++)
	    params[i] = Float_val (Field(Field(param,1),i));
	glLightModelfv (GL_LIGHT_MODEL_AMBIENT, params);
	break;
    case MLTAG_two_side:
	glLightModelf (GL_LIGHT_MODEL_TWO_SIDE,
		       (GLfloat) Int_val(Field(param, 1)));
	break;
    }
    return Val_unit;
}

ML_1 (glLineWidth, Float_val)
ML_0 (glLoadIdentity)
ML_1 (glLoadMatrixf, Float_raw)

ML_1 (glLogicOp, GLenum_val)

CAMLprim value ml_glMaterial (value face, value param)  /* ML */
{
    float params[4];
    int i;

    switch (Field(param,0))
    {
    case MLTAG_shininess:
	params[0] = Float_val (Field(param, 1));
	break;
    default:
	for (i = 0; i < 4; i++)
	    params[i] = Float_val (Field(Field(param, 1), i));
	break;
    }
    glMaterialfv (GLenum_val(face), GLenum_val(Field(param,0)), params);
    return Val_unit;
}

ML_1 (glMatrixMode, GLenum_val)
ML_1 (glMultMatrixf, Float_raw)

ML_3 (glNormal3f, Float_val, Float_val, Float_val)

ML_3 (glOrthof, Pair(arg1,Float_val,Float_val),
      Pair(arg2,Float_val,Float_val), Pair(arg3,Float_val,Float_val))

ML_1 (glPixelStorei, Pair(arg1,GLenum_val,Int_val))

ML_1 (glPointSize, Float_val)
ML_2 (glPolygonOffset, Float_val, Float_val)
ML_0 (glPopMatrix)

ML_0 (glPushMatrix)

CAMLprim value ml_glReadPixels(value x, value y, value w, value h, value format , value raw)  /* ML */
{
  glPixelStorei(GL_PACK_ALIGNMENT, 1); /* XXX why not put back? */
  glReadPixels(Int_val(x),Int_val(y),Int_val(w),Int_val(h),GLenum_val(format),
	       Type_void_raw(raw));
  return Val_unit;
}

ML_bc6 (ml_glReadPixels)
ML_4 (glRotatef, Float_val, Float_val, Float_val, Float_val)
ML_3 (glScalef, Float_val, Float_val, Float_val)

ML_4 (glScissor, Int_val, Int_val, Int_val, Int_val)
ML_1 (glShadeModel, GLenum_val)
ML_3 (glStencilFunc, GLenum_val, Int_val, Int_val)
ML_1 (glStencilMask, Int_val)
ML_3 (glStencilOp, GLenum_val, GLenum_val, GLenum_val)

CAMLprim value ml_glTexEnv (value param)
{
    value params = Field(param,1);
    GLfloat color[4];
    int i;

    switch (Field(param,0)) {
    case MLTAG_mode:
	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GLenum_val(params));
	break;
    case MLTAG_color:
	for (i = 0; i < 4; i++) color[i] = Float_val(Field(params,i));
	glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, color);
	break;
    }
    return Val_unit;
}

CAMLprim value ml_glTexImage2D (value level, value internal,
                                value width, value height, value border,
                                value format, value data)
{
    /* Note: OpenGL ES says the internal format must be the same as the
     * format.  However, Apple's implementation of BGRA currently
     * requires GL_RGBA for the internal format and GL_BGRA for the
     * format.  In fact, it makes sense to allow the internal formats to
     * be a subset of the external ones.
     */
    /* printf("l=%d,i=%x,w=%d,h=%d,b=%d,f=%x,t=%x,d=%x\n", */
    GLenum internal_enum, format_enum;

    switch(internal) {
#ifdef GL_BGRA
    case MLTAG_bgra:
        internal_enum = GL_BGRA;
        break;
#endif
    default:
        internal_enum = GLenum_val(internal);
        break;
    }

    switch(format) {
#ifdef GL_BGRA
    case MLTAG_bgra:
        format_enum = GL_BGRA;
        break;
#endif
    default:
        format_enum = GLenum_val(format);
        break;
    }

    glTexImage2D (GL_TEXTURE_2D,
		  Int_val(level), internal_enum, Int_val(width),
		  Int_val(height), Int_val(border), format_enum,
		  Type_raw(data), Void_raw(data));
    /* fflush(stdout); */
    return Val_unit;
}

ML_bc7 (ml_glTexImage2D)

CAMLprim value ml_glTexSubImage2D (value level, value xoffset, value yoffset,
                                value width, value height,
                                value format, value data)
{
    GLenum format_enum;

    switch(format) {
#ifdef GL_BGRA
    case MLTAG_bgra:
        format_enum = GL_BGRA;
        break;
#endif
    default:
        format_enum = GLenum_val(format);
        break;
    }

    glTexSubImage2D (GL_TEXTURE_2D,
		  Int_val(level), Int_val(xoffset), Int_val(yoffset),
                  Int_val(width), Int_val(height),
                  format_enum, Type_raw(data), Void_raw(data));
    return Val_unit;
}

ML_bc7 (ml_glTexSubImage2D)

CAMLprim value ml_glCompressedTexImage2D (value level, value internal,
                                value width, value height, value border,
                                value data)
{
    GLenum internal_enum;

    switch (internal) {
#ifdef GL_IMG_texture_compression_pvrtc
    case MLTAG_rgb_pvrtc_4bppv1:
        internal_enum = GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG;
        break;
    case MLTAG_rgb_pvrtc_2bppv1:
        internal_enum = GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG;
        break;
    case MLTAG_rgba_pvrtc_4bppv1:
        internal_enum = GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG;
        break;
    case MLTAG_rgba_pvrtc_2bppv1:
        internal_enum = GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG;
        break;
#endif
    default:
        internal_enum = GLenum_val(internal);
        break;
    }

    /*  printf("l=%d,i=%x,w=%d,h=%d,b=%d,s=%d,d=%p\n", */
    glCompressedTexImage2D (GL_TEXTURE_2D,
		  Int_val(level), internal_enum, Int_val(width),
		  Int_val(height), Int_val(border),
		  Int_val(Size_raw(data)), Void_raw(data));
    /*  fflush(stdout); */
    return Val_unit;
}

ML_bc6 (ml_glCompressedTexImage2D)

CAMLprim value ml_glTexParameter (value target, value param)
{
    GLenum targ = GLenum_val(target);
    GLenum pname = GLenum_val(Field(param,0));
    value params = Field(param,1);

    switch (pname) {
    case GL_GENERATE_MIPMAP:
        glTexParameteri (targ, pname, Int_val(params));
        break;
    default:
	glTexParameteri (targ, pname, GLenum_val(params));
	break;
    }
    return Val_unit;
}
ML_2 (glGenTextures, Int_val, (GLuint *) Int_raw)
ML_2 (glBindTexture, GLenum_val, Nativeint_val)

CAMLprim value ml_glDeleteTexture (value texture_id)
{
    GLuint id = Nativeint_val(texture_id);
    glDeleteTextures(1,&id);
    return Val_unit;
}

ML_3 (glTranslatef, Float_val, Float_val, Float_val)

ML_4 (glViewport, Int_val, Int_val, Int_val, Int_val)
